/** XPRINT.C ****************************************************
* Purpose:         Demonstrate calls to print.com
* Author:          John Rex
* Compiler:        Microsoft C 6.0, TC++ 1.0
* Memory model:    Any
* Compile switches: DRIVER - if defined a test driver is included
* Compile with: This routine uses a file normalization routine
*               originally described in V4N3 of The Gazette. The
*               code can be found in that issue or in the source
*               code archives for this issue.
*
* Usage: The test driver is invoked with the command line:
*
*          xprint command [command command ...]
*
*  where command may be any of:
*
*          ?           - status report
*          +file.ext   - add file.ext to files to print
*          -file.ext   - remove file.ext from files to print
*          --          - remove all files from files to print
*
* Source code may be freely used if source is acknowledged.
* Object code may be freely used.
****************************************************************/
#define DRIVER

#include <dos.h>
#include <stdlib.h>

#define SPOOL_INT    0x2F       /* DOS interrupt  */
#define GET_STATE    0x0100     /* function codes */
#define ADD_FILE     0x0101
#define REMOVE_FILE  0x0102
#define CANCEL_ALL   0x0103
#define STATUS_HOLD  0x0104
#define RELEASE_HOLD 0x0105

#define BUFSIZE 100             /* size of file name buffers */

/* A routine to normalize file names by adding drives and a 
** complete path. This routine is taken from C Gazette, Vol. 4,
** No. 3 (1990), and is discussed in that issue. The original 
** version was modified to use \ as delimiter in final form and 
** to return normalized file name in upper rather than lower case
*/
char *normal_name(char *, char *);

/* Macros to extract offset and segment from a FAR pointer. 
** These are used to avoid problems with MSC's versions under 
** some circumstances. */
#define FP_GETOFF(p)  ((unsigned) (p))
#define FP_GETSEG(p)  ((unsigned) ((unsigned long)(p) >> 16) )

#pragma page
#pragma subtitle("Subroutines")
/* Is print.com installed?  Returns 0 if no, 1 if yes. */
int print_installed(void)
{
    union REGS in, out;

    in.x.ax = GET_STATE;
    int86(SPOOL_INT, &in, &out);

    if (out.x.cflag)       return 1;   /* Carry set: failure */

    if (out.h.al == 0xff)  return 1;   /* Is installed       */
    else                   return 0;   /* Not installed      */
}

/* Report status: Is print.com installed? If so, return pointer
** to queue and error count. If call is successful, you MUST
** call release_status after examining queue. The pointer in
** flist will be a FAR pointer to a series of 64-byte, null-
** terminated strings containing the names of the files in the 
** print queue. The last entry in this table will be a string 
** of zero length. See main() for an example of scanning this 
** list of file names.
**/
int get_status(char far **flist, unsigned *error_count)
{
    union REGS in, out;
    struct SREGS sregs;

    in.x.ax = STATUS_HOLD;
    int86x(SPOOL_INT, &in, &out, &sregs);

    if (out.x.cflag) return out.x.ax; /* Carry set: failure */

    *error_count = out.x.dx;
    *flist = (char far *)((((long) sregs.ds) << 16) + out.x.si);

    return 0;   /* return success */
}

/* Release queue so that printing can continue. */
void release_status(void)
{
    union REGS in, out;

    in.x.ax = RELEASE_HOLD;
    int86(SPOOL_INT, &in, &out);
}

/* Add a file to print queue */
int add_file(char *fname)
{
    union REGS in, out;
    struct SREGS sregs;
    char packet[5], work[BUFSIZE];

    /* First, convert the file name to a normalized format */
    if (normal_name(fname, work) == NULL)
        return 3;   /* 3 is error code for PATH NOT FOUND */

    /* Must build up a 5-byte packet of data for DOS. First 
    ** byte should be zero, and the next four bytes are a 
    ** segment:offset pointer to the name of the file to be 
    ** added. A structure was not used in order to avoid problems
    ** with structure padding under different compilers. By 
    ** working with an array, precise control is obtained.
    */
    packet[0] = 0;
    *((unsigned *)(&packet[1])) = FP_GETOFF(((void far *)work));
    *((unsigned *)(&packet[3])) = FP_GETSEG(((void far *)work));

    /* Now, set up the request, passing the address 
    ** of the packet to DOS. */
    in.x.ax = ADD_FILE;
    in.x.dx = FP_GETOFF(  ((void far *)packet) );
    sregs.ds = FP_GETSEG( ((void far *)packet) );
    int86x(SPOOL_INT, &in, &out, &sregs);
    if (out.x.cflag)
        return out.x.ax; /* Carry set: failure */
    else
        return 0;
}

/* Remove a file from print queue. Name may contain wildcards */
int remove_file(char *fname)
{
    union REGS in, out;
    struct SREGS sregs;
    char work[BUFSIZE];

    /* First, normalize file name. Also, since
    ** DOS may expand file name if it contains wildcards, the
    ** the file name must be in a large enough buffer.
    */
    if (normal_name(fname, work) == NULL)
        return 3;   /* 3 is error code for PATH NOT FOUND */

    in.x.ax = REMOVE_FILE;
    in.x.dx =  FP_GETOFF( ((void far *)work) );
    sregs.ds = FP_GETSEG( ((void far *)work) );
    int86x(SPOOL_INT, &in, &out, &sregs);
    if (out.x.cflag)
        return out.x.ax; /* Carry set: failure */
    else
        return 0;
}

/* Clear the print queue. */
int remove_all(void)
{
    union REGS in, out;

    in.x.ax = CANCEL_ALL;
    int86(SPOOL_INT, &in, &out);
    if (out.x.cflag)
        return out.x.ax; /* Carry set: failure */
    else
        return 0;
}

#pragma page()
#pragma subtitle("Driver")
#if defined(DRIVER)
#include <stdio.h>
void main(int argc, char **argv)
{
    char *s;
    int loop, error_code;

    if (argc < 2) 
    {
        printf("Usage: xprint command [command command ...]\n"
        "   Where command may be any of:\n"
        "\t ?         - status report\n"
        "\t +file.ext - add file.ext to files to print\n"
        "\t -file.ext - remove file.ext from files to print\n"
        "\t --        - remove all files from files to print\n");
        return;
    }

    if (!print_installed()) 
    {
        printf("Print.com is not installed.\n");
        return;
    }
    printf("Print.com is installed.\n");

    for (loop=1; loop<argc; loop++) 
    {
		s=argv[loop];

        /* report status */
        if (*s == '?') 
        {
            char far *flist;
            unsigned error_count;

            error_code = get_status(&flist, &error_count);
            if (error_code)
                perror("Error while checking status");
            else if (flist == NULL)
                printf("Could not obtain status information.\n");
            else {
                printf("%u errors so far.", error_count);
                if (*flist == '\0')
                    printf(" Print queue is empty.\n");
                else 
                {
                    printf(" Print queue contains:\n");
            /*
            **  N.B.: flist is a FAR pointer, and must be handled
            **   carefully, depending on your memory model! Here
            **   we take a general-purpose approach that will
            **   work under any memory model.
            **/
                    while (*flist != '\0') 
                    {
                        char far *q;
                        fputchar(' ');
                        for (q=flist; *q; q++)
                            fputchar(*q);
                        fputchar('\n');
                        flist += 64;
                    }
                }
            }
            release_status();
        }

        /* add a file to print queue */
        else if (*s == '+') 
        {
            s++;
            error_code = add_file(s);
            if (error_code)
                perror("Error while adding file to queue");
            else
                printf(" Added file %s to queue.\n", s);
        }

        /* remove one or more files from print queue */
        else if (*s == '-') 
        {
            s++;
            if (*s == '-') 
            {
                error_code = remove_all();
                if (error_code)
                    perror("Error while clearing queue");
                else
                    printf(" Print queue cleared.\n");
            }
            else 
            {
                error_code = remove_file(s);
                if (error_code)
                   perror(
                    "Error while removing file from queue");
                else
                   printf(" Removed file %s from queue.\n", s);
            }
        }
        else printf("Unknown option: %s\n", s);
    }
}
#endif